/* Bediengert fr Herd

 Realisiert mit ATmega324A
 16MHz Quarz

 Version 2.3    2013-04-04 6:36
 
*/

/* Todo
 
 Ideen: 
   - DrehgeberDruck schaltet auf Timer um. Langer Druck macht H+ oder V+
   - Drehgeberdruck: Platte umschalten 0% / 100%
   - Wenn Platte aus: 2 Incs bis auf 1  



   Die Bedienung ist in mehreren Ebenen (Normal/Timer/Ofen/Schaltuhr/Service) organisiert.
   Fr jeden der 4 Drehgeber wird je Ebene ein Zhlerstand verwaltet.
   Drehen am Drehgeber ndert den Zhlerstand um einen Inc-Wert, begrenzt durch Min/Max/Modulo Werte.
   Die Drehgeberleitungen werden via PinChangeInterrupt berwacht, die Tasten alle 5ms gepollt.

   Die Timerwerte werden in Sekunden verwaltet.

*/


//#define F_CPU 16000000UL    /* CPU clock in Hertz */

#include <util/delay.h> 
#include <avr/io.h>
#include <avr/interrupt.h>
#include <string.h> //memset
#include <avr/pgmspace.h>
#include "def.h"

FUSES = {0xF7, 0xD9, 0xFF}; //Lo Hi Ext

volatile T_global  g;
U8 txt[7];  //Universell zum Senden genutzter Buffer

//-----------------------------------------------------------------------------
void initHW(void){
//-----------------------------------------------------------------------------

  //Globale var auf 0
  memset((U8 *)(&g), 0, sizeof(T_global));  

  MCUCR = (1<<JTD); //JTAG aus, sonst geht PortC nicht!
  PRR0 = (1<<PRUSART1) | (1<< PRTWI) | (1<<PRSPI) | (1<<PRADC); //Power reduction
                     
  DDRA = 0;
  DDRB = 0;
  DDRC = 0xff;
  DDRD = BIT1 | BIT6 | BIT7;

  //Pull-up Widerstnde einschalten 
  PORTA = 0xff; //Alle
  PORTB = BIT0 | BIT1 | BIT2 | BIT3 | BIT4;  //Mein DCF77-Empfnger braucht einen Pull-Up an PB4
  PORTD = BIT2 | BIT3;

  PORTC = 0; 
  
  PCICR |=(1<<PCIE0);  // enables range PCINT0..7        (Drehgeber)
  PCMSK0=0xff;         // enables each pin of PCINT0..7
  
  EIMSK = 0;  //Keine externen Interrupts
  
  //Timer0: 1KHz Takt
  TCCR0A = (1<<WGM01); //CTC Mode(2),
  TCCR0B = (1<<CS01) | (1<<CS00); // clk/64  (250KHz bei 16MHz Takt)
  OCR0A = 249;
  TIMSK0 = (1<<OCIE0A);

  //Timer1: 10Hz Takt
  TCCR1A = 0;
  TCCR1B = (1<<WGM12) | (1<<CS11)| (1<<CS10); //CTC Mode(4), clk/64  (250KHz bei 16MHz Takt)
  OCR1A = 24999;
  TIMSK1 = (1<<OCIE1A);

  //Timer0: Beeper (1-4KHz) wird nach Init wieder ausgeschaltet
  TCCR2A = (1<<WGM21) | (1<<COM2A0); //CTC Mode(2), toggle OC2A output
  TCCR2B = (1<<CS22) ; // clk/64  (250KHz bei 16MHz Takt)
  OCR2A = 250;// 
  TIMSK2 = 0;
}

//-----------------------------------------------------------------------------
void incCounter(U8 kanal){
//-----------------------------------------------------------------------------
  g.bChange=1;
  g.iCounter[g.bEbene][kanal] += g.iInc[g.bEbene][kanal];
  if (g.iCounter[g.bEbene][kanal] > g.iMax[g.bEbene][kanal]) g.iCounter[g.bEbene][kanal] = g.iMax[g.bEbene][kanal];
  if (g.iCounter[g.bEbene][kanal] < g.iMin[g.bEbene][kanal]) g.iCounter[g.bEbene][kanal] = g.iMin[g.bEbene][kanal];
  g.iCounter[g.bEbene][kanal] %= g.iModulo[g.bEbene][kanal];
}

//-----------------------------------------------------------------------------
void decCounter(U8 kanal){
//-----------------------------------------------------------------------------
  g.bChange=1;
  g.iCounter[g.bEbene][kanal] -= g.iInc[g.bEbene][kanal];
  if (g.iCounter[g.bEbene][kanal] > g.iMax[g.bEbene][kanal]) g.iCounter[g.bEbene][kanal] = g.iMax[g.bEbene][kanal];
  if (g.iCounter[g.bEbene][kanal] < g.iMin[g.bEbene][kanal]) g.iCounter[g.bEbene][kanal] = g.iMin[g.bEbene][kanal];
  g.iCounter[g.bEbene][kanal] = (g.iCounter[g.bEbene][kanal] + g.iModulo[g.bEbene][kanal]) % g.iModulo[g.bEbene][kanal];
}

//-----------------------------------------------------------------------------
void drehgeber(U8 a, U8 b, U8 kanal){
//-----------------------------------------------------------------------------
// A:Puls B: Richtung 
  static U8 bisher[4];

  if (bisher[kanal] != a){ //Puls entdeckt
    bisher[kanal] = a;
    if (!b){ //Richtung bewerten
      if (a) incCounter(kanal); else decCounter(kanal);
    }else{
      if (a) decCounter(kanal); else incCounter(kanal);
    }
  }
} 
//-----------------------------------------------------------------------------
ISR(PCINT0_vect){ //Drehgeber
//-----------------------------------------------------------------------------
  drehgeber(DREH1_A, DREH1_B, 0);
  drehgeber(DREH2_A, DREH2_B, 1);
  drehgeber(DREH3_A, DREH3_B, 2);
  drehgeber(DREH4_A, DREH4_B, 3);
}   


#if 0

#if 0
//-----------------------------------------------------------------------------
ISR(TIMER0_COMPA_vect){ //Piezo-Beeper
//-----------------------------------------------------------------------------
  PIND |= BIT7; //Port umschalten
}
#else  //handoptimiert in Assembler
    void  __attribute__ ((naked)) TIMER0_COMPA_vect(void){
      asm("sbi    0x09, 7"  );    //Port umschalten
      asm("reti"  );    //Status sichern
    }
#endif
#endif
//-----------------------------------------------------------------------------
ISR(TIMER1_COMPA_vect){ //100ms Timer 
//-----------------------------------------------------------------------------
static U8 t10=10;
  t10--;
  if (0==t10){
    t10 = 10;
    g.bToggle1000ms = !g.bToggle1000ms;
  }
}
//-----------------------------------------------------------------------------
ISR(TIMER0_COMPA_vect){ //Millisekunden Timer 
//-----------------------------------------------------------------------------
  g.iTimeTic1ms++;
  if (31000 == g.iTimeTic1ms) g.iTimeTic1ms=0;
}




//--------------------------------------------------------
ISR(USART0_RX_vect){ //Zeichen von RS232 empfangen
//--------------------------------------------------------
  g.RxBuffer[g.RxBufferWrite++] = UDR0;
  if (RX_BUFFERSIZE == g.RxBufferWrite){
    g.RxBufferWrite=0;
  }
}

//-----------------------------------------------------------------------------
void beeperEin(void){
//-----------------------------------------------------------------------------
  TCCR2A = (1<<WGM21) | (1<<COM2A0); //CTC Mode(2), toggle OC2A output
  TCCR2B = (1<<CS22) ; // clk/64  (250KHz bei 16MHz Takt)
  OCR2A = 254;
  g.mindestPiepsTimer = 40;  //*20ms=800ms
}

//-----------------------------------------------------------------------------
void beeperAus(void){
//-----------------------------------------------------------------------------
  TCCR2A = 0;
  TCCR2B = 0;
  PORTD &=~ BIT7; //Transistor aus
}

/*
//--------------------------------------------------------
U8 RxDataAvail(void){ //Wieviele Zeichen sind im Puffer
//--------------------------------------------------------
  return (g.RxBufferWrite + RX_BUFFERSIZE - g.RxBufferRead) % RX_BUFFERSIZE;
}
*/

/**************************************************************************/
/* Zeitdifferenz mit 1ms Granularitaet */
/* Laufzeit bis einschl. 30s zulaessig */
/**************************************************************************/
I16 deltaT_ms(I16 startZeit) {
  I16 delta;
  cli();
  delta = g.iTimeTic1ms;
  sei();
  delta -= startZeit;
  if (delta < 0) delta += 31000;
  return delta;
}
//--------------------------------------------------------
void USART_Init(void){
//--------------------------------------------------------
 
  /* Set baud rate: round(f_osc / (8*baudrate)) - 1  wenn U2X=1*/

  UBRR0H = 0;
  UBRR0L = 207; //9600 Baud bei 16MHz

  UCSR0A=(1<<U2X0); //set U2X

  /* Set frame format:          1stop bit   8data,*/
  UCSR0C =  (3<<UCSZ00);

  // Tx + RX enable. RX interrupt
  UCSR0B = (1<<TXEN0)|(1<<RXEN0)|(1<<RXCIE0);
}

//--------------------------------------------------------
void USART_Transmit(U8 data ){
//--------------------------------------------------------
  /* Wait for empty transmit buffer */
  while ( !( UCSR0A & (1<<UDRE0)) )
  ;
  /* Put data into buffer, sends the data */
  UDR0 = data;
}


//-----------------------------------------------------------------------------
void sendTxt(U8 *txt){
//-----------------------------------------------------------------------------
  while (*txt){
    USART_Transmit(*txt++);
  }
}

//-----------------------------------------------------------------------------
void anheizVsenden(U8 on){
//-----------------------------------------------------------------------------
  txt[0]='*';
  txt[1]='v';
  txt[2]=48+on;
  txt[3]=48+on;
  txt[4]=13;
  txt[5]=0;
  sendTxt(txt);
}
//-----------------------------------------------------------------------------
void anheizHsenden(U8 on){
//-----------------------------------------------------------------------------
  txt[0]='*';
  txt[1]='h';
  txt[2]=48+on;
  txt[3]=48+on;
  txt[4]=13;
  txt[5]=0;
  sendTxt(txt);
}

//-----------------------------------------------------------------------------
void sollwertSenden(U8 n, U8 on, U8 off){ //n=Kanal 1..7
//-----------------------------------------------------------------------------
  txt[0]='*';
  txt[1]=48+n;
  txt[2]=48+on;
  txt[3]=48+off;
  txt[4]=13;
  txt[5]=0;
  sendTxt(txt);
}

//-----------------------------------------------------------------------------
void sollwerteSenden(void){
//-----------------------------------------------------------------------------
  static U8 onVal [15]={0, 1, 1, 1, 2, 3, 4, 3, 1, 4, 2, 3, 4, 9, 9};
  static U8 offVal[15]={0,20,10, 5, 7, 7, 7, 4, 1, 3, 1, 1, 1, 1, 0};
  //0 4 9 16 22 30 36 42 50 57 66 75 80 90 100 %

  static U8 onValHalo [15]={0, 1, 1, 1, 1, 1, 1, 2, 1, 3, 2, 3, 4, 9, 9};
  static U8 offValHalo[15]={0,20,10, 5, 4, 3, 2, 3, 1, 2, 1, 1, 1, 1, 0};
//  0 4 9 16 20 25 33 40 50 60 66 75 80 90 100


  static int anheizV=0;
  static int anheizH=0;
//  static int anheizVms=-1;
//  static int anheizHms=-1;
  static int iRelaisaenderung;
  static int merkHeiz=0;
  static int merkTemp=0;
  static U8  merkLicht=0;
  static U8  kuehloffset=2;   //fr Hysterese der Gehuselftung
  U8 bOfenRelais;
  U8 bUmluft;
  
  /* Kanle:
      1= VL innen
      2= VL aussen
  
      3= HL 
      4= HL lang
  
      5= VR1
      6= VR2
      7= HR
  
  */
#if 1  // 0 fr Debug Ofenrelais
  
  //--VL--
  sollwertSenden(1,onVal[g.iAusgabewert[KNOPF_UL]], offVal[g.iAusgabewert[KNOPF_UL]]);
  if (g.bAussenV){
    sollwertSenden(2,onVal[g.iAusgabewert[KNOPF_UL]], offVal[g.iAusgabewert[KNOPF_UL]]);
  }else{
    sollwertSenden(2,0,9);
  }

  //--HL--
  sollwertSenden(3,onVal[g.iAusgabewert[KNOPF_OL]], offVal[g.iAusgabewert[KNOPF_OL]]);
  if (g.bAussenH){
    sollwertSenden(4,onVal[g.iAusgabewert[KNOPF_OL]], offVal[g.iAusgabewert[KNOPF_OL]]);
  }else{
    sollwertSenden(4,0,9);
  }


  //--HR--
  // Halogenlampen erst langsam erwrmen. Daher bei kalter Lampe nur 1ms Pulse draufgeben
  if (g.iAusgabewert[KNOPF_OR]){
    anheizH++;
    if (anheizH > 9) anheizH=9;
    if (anheizH <5){
      anheizHsenden(1);
      sollwertSenden(7,2,0);
    }else{
      sollwertSenden(7,onValHalo[g.iAusgabewert[KNOPF_OR]], offValHalo[g.iAusgabewert[KNOPF_OR]]);
      anheizHsenden(0);
    }
  }else{
    anheizH=0;
    sollwertSenden(7,0,9);
  }
  
  //--VR--
  // Halogenlampen erst langsam erwrmen. Daher bei kalter Lampe nur 1ms Pulse draufgeben
  if (g.iAusgabewert[KNOPF_UR]){
    anheizV++;
    if (anheizV > 9) anheizV=9;
    if (anheizV<3){
      anheizVsenden(1);      //1ms Pulse 
      sollwertSenden(9,2,0); //Interleave
    }else      
    if (anheizV <5){
      anheizVsenden(1);        //1ms Pulse
      sollwertSenden(5,2,0);   //Kein Interleave
      sollwertSenden(6,2,0);
    }else{
      sollwertSenden(5,onValHalo[g.iAusgabewert[KNOPF_UR]], offValHalo[g.iAusgabewert[KNOPF_UR]]);
      sollwertSenden(6,onValHalo[g.iAusgabewert[KNOPF_UR]], offValHalo[g.iAusgabewert[KNOPF_UR]]);
      anheizVsenden(0);
    }
  }else{
    anheizV=0;
    sollwertSenden(5,0,9);
    sollwertSenden(6,0,9);
  }
  
  
  /*
Herdrelais:
   1 Relais Oberhitze (uerer Heizrahmen)
   2 Relais Grill     (innere Heizmander) <-ggf. OH mit dazu nehmen
   4 Relais Unterhitze
   8 Relais Umlufthitze
  16 Relais Umluftmotor aus (ACHTUNG: invertiert!) 
  32 Relais Gehusekhlung invertieren
            (Ist bei allen Heizungen mit an. Wenn Heizungen aus kann damit zugeschalktet werden)
  
    "-AUS-",   //0
    "Licht",
    " Um  ",   //2
    "Um+Uh",   
    "Um+Oh",   //4
    " Oh  ",      
    " Uh  ",   //6
    "Oh+Uh",   
    "Grill"    //8


*/
  bOfenRelais=0;
  bUmluft=0;
  if (g.bFreigabeOfen){
    //Sollwert zu niedrig, dann Heizung ein
    if (g.iOfentempNormiert < g.iCounter[EBENE_OFEN][0]){
      switch (g.iCounter[EBENE_OFEN][1]){
        case 2: bOfenRelais= 8;        break; //Um
        case 3: bOfenRelais= 8+ 4;     break; //Um+Uh
        case 4: bOfenRelais= 8+ 1;     break; //Um+Oh
        case 5: bOfenRelais= 1+16;  break; //Oh
        case 6: bOfenRelais= 4+16;  break; //Uh
        case 7: bOfenRelais= 1+4+16;break; //Oh+Uh 
        case 8: bOfenRelais= 2+16;  break; //Grill
      }
    }
    if ((g.iCounter[EBENE_OFEN][1] > 1) 
     && (g.iCounter[EBENE_OFEN][1] < 5) ){  //2,3,4
       bUmluft=1;    
    }

  }

  if (!bUmluft){
     bOfenRelais |= 16; //Umluftmotor aus 
  }
  //ggf. Khlung ein
  if ((g.iOfentempCelsius > 50+kuehloffset) || (bOfenRelais &~16) || (g.iPlatinentempRoh < 100) || bUmluft){
    bOfenRelais|=32; //Gehusekhlung ein
    kuehloffset=0;
  }else{
    kuehloffset=2;
  }

  //nderungen nur gelegentlich senden (alle 5s)
  if (
       (merkHeiz!=g.iCounter[EBENE_OFEN][1]) ||
       (merkTemp!=g.iCounter[EBENE_OFEN][0])
  ){
    merkHeiz=g.iCounter[EBENE_OFEN][1];
    merkTemp=g.iCounter[EBENE_OFEN][0];
    iRelaisaenderung=deltaT_ms(0);
  }

  if ((deltaT_ms(iRelaisaenderung) > 3000) || (merkLicht != (g.iCounter[EBENE_OFEN][1] > 0)?1:0)){
    merkLicht = (g.iCounter[EBENE_OFEN][1] > 0)?1:0;
    sollwertSenden(17, 48+bOfenRelais, merkLicht && g.bFreigabeOfen); //(3.Parameter=Licht)
    iRelaisaenderung=deltaT_ms(0);
  }
#else //Debug Ofenrelais
  if (deltaT_ms(iRelaisaenderung) > 500){
    sollwertSenden(1,0,9);
    sollwertSenden(2,0,9);
    sollwertSenden(3,0,9);
    sollwertSenden(4,0,9);
    sollwertSenden(5,0,9);
    sollwertSenden(6,0,9);
    
    bOfenRelais = (1 << g.iCounter[EBENE_NORMAL][1]) >>1;
    bOfenRelais |= (1 << g.iCounter[EBENE_NORMAL][3]) >>1;
    sollwertSenden(17, 48+bOfenRelais, merkLicht && g.bFreigabeOfen); //(3.Parameter=Licht)
    iRelaisaenderung=deltaT_ms(0);
  }
#endif

}
//-----------------------------------------------------------------------------
void filterSchritt(void){
//-----------------------------------------------------------------------------
  int i;
  int delta;
  for (i=0; i<4; i++){
    if (g.iCounter[EBENE_NORMAL][i]*1000 < g.iSpeicher[i]) {
      g.iSpeicher[i] -= 1 + (g.iMax[EBENE_NORMAL][i] - g.iAusgabewert[i]) / 3; 
    }
    if (g.iCounter[EBENE_NORMAL][i]*1000 > g.iSpeicher[i]){
      g.iSpeicher[i] += g.iAusgabewert[i]; 
    }
    delta=g.iCounter[EBENE_NORMAL][i]*1000 - g.iSpeicher[i];
    g.iAusgabewert[i] = g.iCounter[EBENE_NORMAL][i] + delta / g.iUeberhoehung[i];
    if (g.iAusgabewert[i] >= g.iMax[EBENE_NORMAL][i]){
      g.iAusgabewert[i] = g.iMax[EBENE_NORMAL][i];
    }
    if (g.iAusgabewert[i] < 0){
      g.iAusgabewert[i] = 0;
    }
  }    
}

//-----------------------------------------------------------------------------
void clearKeys(void){
//-----------------------------------------------------------------------------
  g.bTaste1=0;
  g.bTaste2=0;
  g.bTaste3=0;
  g.bTaste4=0;
  g.bTasteOfen=0;
  g.bTasteTimer=0;
  g.bEineTaste=0;
}


//-----------------------------------------------------------------------------
void ebeneNormal(void){
//-----------------------------------------------------------------------------
  if (g.bChange){
    g.bChange=0;
    g.iIdleMs=deltaT_ms(0);
    g.bSendeSollwerte=6;
    g.bRefreshDisplay=1;
    g.bLichtCounter=0;
  }    
  if (!(TASTE_2 || TASTE_4)){ //Service Menu
    clearKeys();
    g.bEbene=EBENE_SERVICE;
    g.iIdleMs=deltaT_ms(0);
    g.bRefreshDisplay=1;
    return;
  }
    
  if (g.bTasteTimer || g.bTaste2 || g.bTaste4){
    clearKeys();
    g.bEbene=EBENE_TIMER;
    g.iIdleMs=deltaT_ms(0);
    g.bRefreshDisplay=1;
  }
  if (g.bTasteOfen){
    clearKeys();
    g.bEbene=EBENE_OFEN;
    g.iIdleMs=deltaT_ms(0);
  }
  if (g.bTaste1){
    g.bAussenH = !g.bAussenH;
    clearKeys();
    g.bRefreshDisplay=1;
  }
  if (g.bTaste3){
    g.bAussenV = !g.bAussenV;
    clearKeys();
    g.bRefreshDisplay=1;
  }
}

//-----------------------------------------------------------------------------
void ebeneTimer(void){
//-----------------------------------------------------------------------------
  if (g.bChange){
    g.bChange=0;
    g.bRefreshDisplay=1;
    g.iIdleMs=deltaT_ms(0);
  }
  
  //Timeout
  if (deltaT_ms(g.iIdleMs) > 5000){
    g.bEbene=EBENE_NORMAL;
    clearKeys();
    g.bRefreshDisplay=1;    
  }

  if (!(TASTE_2 || TASTE_4)){ //Service Menu
    clearKeys();
    g.bEbene=EBENE_SERVICE;
    g.iIdleMs=deltaT_ms(0);
    return;
  }

  //Wenn irgendein Knopf gedrckt
  if (g.bEineTaste){
    clearKeys();
    g.bEbene=EBENE_NORMAL;    
    g.bRefreshDisplay=1;    
  }

}

//-----------------------------------------------------------------------------
void ebeneOfen(void){
//-----------------------------------------------------------------------------
  if (g.bChange){
    g.bChange=0;
    g.bRefreshDisplay=1;
    g.iIdleMs=deltaT_ms(0);
  }
  
  //Timeout
  if (deltaT_ms(g.iIdleMs) > 8000){
    clearKeys();
    g.bEbene=EBENE_NORMAL;  
    g.bRefreshDisplay=1;    
  }
  //Wenn  Drehknopf 1..3 gedrckt
  if (g.bTaste1 || g.bTaste2 || g.bTaste3){
    clearKeys();
    g.bEbene=EBENE_NORMAL;    
    g.bRefreshDisplay=1;    
  }
  if (g.bTasteOfen){
    clearKeys();
    g.bEbene=EBENE_NORMAL;
  }
  if (g.bTasteTimer){
    g.bEbene=EBENE_SCHALTUHR;
    clearKeys();
    g.bRefreshDisplay=1;    
  }
}

//-----------------------------------------------------------------------------
void ebeneSchaltuhr(void){
//-----------------------------------------------------------------------------
  if (g.bChange){
    g.bChange=0;
    g.bRefreshDisplay=1;
    g.iIdleMs=deltaT_ms(0);
  }
  if (g.bTasteOfen){
    g.bEbene=EBENE_OFEN;
    clearKeys();
  }
  if (g.bTasteTimer){
    g.bEbene=EBENE_OFEN;
    clearKeys();
  }
  
  //Timeout
  if (deltaT_ms(g.iIdleMs) > 8000){
    g.bEbene=EBENE_OFEN;  
    g.bRefreshDisplay=1;    
    g.iIdleMs=deltaT_ms(0);
    clearKeys();
  }
  //Wenn  Drehknopf 1..4 gedrckt
  if (g.bTaste1 || g.bTaste2 || g.bTaste3 || g.bTaste4){
    g.bEbene=EBENE_NORMAL;  
    clearKeys();
    g.iIdleMs=deltaT_ms(0);
  }
}
//-----------------------------------------------------------------------------
void ebeneService(void){
//-----------------------------------------------------------------------------
  if (g.bChange){
    g.bChange=0;
    g.bRefreshDisplay=1;
    g.iIdleMs=deltaT_ms(0);
  }
  
  //Timeout
  if (deltaT_ms(g.iIdleMs) > 20000){
//    g.bEbene=EBENE_NORMAL;  
    g.bRefreshDisplay=1;    
  }
  if (g.bTasteTimer){
    g.bEbene=EBENE_NORMAL;
    g.bTasteTimer=0;
    g.bRefreshDisplay=1;    
  }
}

//-----------------------------------------------------------------------------
void tastenPollen(void){
//-----------------------------------------------------------------------------
  static U8 bTasteTimer=0;
  static U8 bTasteTimerLang=0;
  static U8 bTasteOfen=0;
  static U8 bTimerFlag=0;
  static U8 bTimerFlagLang=0;
  static U8 bOfenFlag=0;
  
  static U8 bFlag1=0;
  static U8 bFlag2=0;
  static U8 bFlag3=0;
  static U8 bFlag4=0;
  
  static U8 bTaste1=0;
  static U8 bTaste2=0;
  static U8 bTaste3=0;
  static U8 bTaste4=0;

  
  //---Tasten abfragen---
  
  //--- Eingang ist Low, wenn Taste gedrckt.

  if (TASTE_OFEN){  //Taste losgelassen
    if (bTasteOfen){
      bTasteOfen--;
    }else{
      bOfenFlag=0; 
    }
  }else{  //Taste gedrckt
    bTasteOfen = 20;
    if (!bOfenFlag){
      bOfenFlag=1;
      g.bChange=1;
      g.bTasteOfen=1;
      g.bEineTaste=1;
    }
  }
  if (TASTE_TIMER){  //Taste losgelassen
    if (bTasteTimer){
      bTasteTimer--;
    }else{
      bTimerFlag=0; 
    }
    bTasteTimerLang=0;
    bTimerFlagLang=0;
    g.bZeigeTonAus=0;
  }else{  //Taste gedrckt
    bTasteTimer = 20;
    if (!bTimerFlag){
      bTimerFlag=1;
      g.bChange=1;
      g.bTasteTimer=1;
      g.bEineTaste=1;
    }
    if (bTasteTimerLang<250){
      bTasteTimerLang++;
    }else{
      if (!bTimerFlagLang){
        bTimerFlagLang=1;
        g.bTasteTimerLang=1;
        g.bZeigeTonAus=1;
        g.bBeeperEnabled = !g.bBeeperEnabled;
      }
    }
  } 


  //Wenn Drehgeber gedrckt: g.bTaste#  auf 1
  
  //Drehgebertaste 1
  if (TASTE_1){  //Taste losgelassen
    if (bTaste1){
      bTaste1--;
    }else{
      bFlag1=0; 
    }
  }else{  //Taste gedrckt
    bTaste1 = 30;    //Sperrzeit:30*5=150ms
    if (!bFlag1){
      g.bTaste1=1;
      g.bEineTaste=1;
      g.bChange=1;
      bFlag1=1;
    }
  }

  //Drehgebertaste 2
  if (TASTE_2){  //Taste losgelassen
    if (bTaste2){
      bTaste2--;
    }else{
      bFlag2=0; 
    }
  }else{  //Taste gedrckt
    bTaste2 = 30;    //Sperrzeit:30*5=150ms
    if (!bFlag2){
      g.bTaste2=1;
      g.bEineTaste=1;
      g.bChange=1;
      bFlag2=1;
    }
  }

  //Drehgebertaste 3
  if (TASTE_3){  //Taste losgelassen
    if (bTaste3){
      bTaste3--;
    }else{
      bFlag3=0; 
    }
  }else{  //Taste gedrckt
    bTaste3 = 30;    //Sperrzeit:30*5=150ms
    if (!bFlag3){
      g.bTaste3=1;
      g.bEineTaste=1;
      g.bChange=1;
      bFlag3=1;
    }
  }

  //Drehgebertaste 4
  if (TASTE_4){  //Taste losgelassen
    if (bTaste4){
      bTaste4--;
    }else{
      bFlag4=0; 
    }
  }else{  //Taste gedrckt
    bTaste4 = 30;    //Sperrzeit:30*5=150ms
    if (!bFlag4){
      g.bTaste4=1;
      g.bEineTaste=1;
      g.bChange=1;
      bFlag4=1;
    }
  }

  if (g.bEineTaste && g.bBlinken && (0==g.mindestPiepsTimer)){
    g.bBlinken=0;
    beeperAus();
    g.bAbgelaufeneTimer=0;
    clearKeys(); //Tastendruck soll nur Beeper+Blinken ausschalten und sonst nix bewirken.
  }

}

//-----------------------------------------------------------------------------
I16 iNormiereKerntemp(I16 roh){
//-----------------------------------------------------------------------------
  static U8 kt []={
    //0   1   2   3   4   5   6   7   8   9
      3, 10, 15, 19, 26, 33, 38, 43, 49, 54, // 0.. 9
     58, 66, 73, 79, 84, 90, 95, 99,105,109, //10..19
    114,120,123,128,133,137,140,144,149,152, //20..29
    156,159,163,166,169,172,175,178,180,182, //30..39
    185,188,190,192,195,197,199,201,204,206, //40..49
    209,210,212,214,216,217,219,221,222,223, //50..59
    224,225,226,228,229,230,231,232,233,234, //60..69
    235,236,236,237,238,239,240,240,241,242, //70..79
    243,244,245,246,247,248,248,249,249,250, //80..89
    250,251,251,252,253,254,                 //90..95 
    255}; // BegrenzungS
  static U8 kern=0;  //Tiefpass zur Erkennung:Fhler eingesteckt
  U8 i;
  U8 r;

  if (roh >253){
    kern = 0;
    g.mitKern = 0;
    return 0;
  }else{
    if (roh < 250){
      kern=10; 
    }else {
      if (kern < 10) kern++;
    }
    if (kern>9) g.mitKern = 1;
    else        g.mitKern = 0;

    if (roh > 253) return 0;
    r = roh;
    i=0;
    //Abkrzungen um Rechenzeit klein zu halten
    if (r > kt[64]) i=65; else 
    if (r > kt[48]) i=49; else
    if (r > kt[32]) i=33; else
    if (r > kt[16]) i=17; 

    while (r > kt[i]){
      i++;
    }    
    return (I16)(95 - i);
  }
}
//-----------------------------------------------------------------------------
I16 iNormiereOfentemp(I16 roh){
//-----------------------------------------------------------------------------
#if 1
  U8 t;
  static I16 ot1 [10]={
  //  0   1   2   3   4   5   6   7   8   9
    270,269,268,267,266,265,264,260,250,240  //0..9 Unter 8: Unbrauchbar; Nicht gemessen!
  };
  PROGMEM static U8 ot2 [246]={
  //  0   1   2   3   4   5   6   7   8   9
    231,224,220,215,210,206,203,201,199,195,  //10..19
    193,191,188,186,185,183,181,179,177,175,  //20..29
    173,172,171,169,168,166,165,163,162,161,  //30..39
    160,159,158,156,155,154,153,152,151,150,  //40..49
    149,148,147,146,145,144,143,142,141,140,  //50..59
    139,138,137,137,136,135,134,133,132,132,  //60..69
    131,131,130,129,128,127,127,126,125,125,  //70..79
    124,123,123,122,121,120,119,119,118,118,  //80..89
    117,116,116,115,115,114,113,112,112,111,  //90..99
    111,110,109,109,108,107,107,106,106,105,  //..109
    104,103,103,102,102,101,101,101,100,100,  //..119
     99, 98, 98, 97, 97, 96, 96, 95, 95 ,94,  //..129
     93, 93, 92, 92 ,91, 91, 90, 90, 89, 89,  //..139
     88, 87, 87, 86, 86, 85, 85, 84, 84, 83,  //..149
     83, 82, 81, 81, 80, 80, 79 ,79, 78, 78,  //..159
     77, 77, 76, 75, 75, 75, 74, 74, 73, 72,  //..169
     72, 71, 71, 70, 70, 70, 69, 69, 68, 67,  //..179
     66, 66, 65, 65, 64, 64, 63, 63, 62, 61,  //..189
     61, 60, 59, 59, 58, 58, 57, 57, 56, 56,  //..199
     55, 54, 53, 53, 52, 52, 51, 51, 50, 49,  //..209
     49, 48, 47, 46, 45, 44, 43, 42, 41, 40,  //..219
     40, 39, 39, 38, 37, 36, 36, 35, 35, 34,  //220..229
     35, 35, 34, 34, 33, 33, 32, 32, 31, 31,//..239
     30, 29, 29, 28, 28, 27, 27, 26, 26, 26,//..249
     25, 24, 23, 22, 21, 21//..255 
     };
     if (roh < 255){
       t = (U8)roh;
       if (t < 10){
         return ot1[t];
       }else{      
         t-=10;
         return pgm_read_byte(ot2 + t);       
//         return (I16)(ot2[t]);
       }  
     }
     return 20;
#else
  static I16 ot []={
      3, 10, 15, 19, 26, 33, 38, 43, 49, 54, //250..223
      58, 66, 73, 79, 84, 90, 95, 99,105,109, //220..193
    114,120,123,128,133,137,140,144,149,152, //190..163
    156,159,163,166,169,172,175,178,180,182, //160..133
    185,188,190,192,195,197,199,201,204,206, //130..103
    209,210,212,214,216,217,219,221,222,223, //100.. 73
    224,225,226,228,229,230,231,232,233,234, // 70.. 43
    235,236,237,238,239,240,240,240,241,242, // 40.. 13
    243,243,244,245,245,246,247,247,248,248, // 10..-17
    249,250,250,250,251,252,
    256};  //Begrenzung
  int i;
  i=0;
  while (roh > ot[i]){
    i++;
  }    
  i*=3;
  if (i > 250) i=250;
  return 250 - i;
#endif    
}
//-----------------------------------------------------------------------------
void rxChar(char c){
//-----------------------------------------------------------------------------
  static U8 fifo[8]={0,0,0,0,0,0,0,0};
  int i;

  for (i=1; i<8; i++){
    fifo[i-1]=fifo[i];
  }
  
  fifo[7] = c;

/*  Im PIC des Herdes:    
      putc('#');
      putc((glob.tempOfen     / 16) + 0x40);
      putc((glob.tempOfen     % 16) + 0x40);
      putc((glob.tempSpiess   / 16) + 0x40);
      putc((glob.tempSpiess   % 16) + 0x40);
      putc((glob.tempPlatine  / 16) + 0x40);
      putc((glob.tempPlatine  % 16) + 0x40);
      putc(13);
*/     


  if (13==fifo[7]){
    if ('#'==fifo[0]){
      g.iOfentempRoh = (fifo[1]-64)*16 + fifo[2]-64;
      g.iOfentempNormiert = iNormiereOfentemp(g.iOfentempRoh);
      g.iKerntempRoh = (fifo[3]-64)*16 + fifo[4]-64;
      g.iKerntempNormiert = iNormiereKerntemp(g.iKerntempRoh);
      g.iPlatinentempRoh = (fifo[5]-64)*16 + fifo[6]-64;
    }  
  }
}


//-----------------------------------------------------------------------------
void task_control(void){  //alle 5ms aufrufen
//-----------------------------------------------------------------------------
static U8   init=1;
static U8   b5msTic=0;
static U8   bMerkTic=0;
static I16  iFilterzeitMs=0;
static I16  sollwertGesendetMs=0;
static I16  iSchaltuhrBeginn;

U8    i;
I16   iJetzt;
I16   iTimerSec;
I16   iTimerSecNext;

  if (init){
    init=0;

//    WDTS.c = 1;    // Den Watchdog fttern
    
    //Zhler Null setzen - Notwendig, da sonst die Platten durch
    //pollen der Drehgeber bereits auf 1 stehen knnten
    g.iCounter[EBENE_NORMAL][0]=0;
    g.iCounter[EBENE_NORMAL][1]=0;
    g.iCounter[EBENE_NORMAL][2]=0;
    g.iCounter[EBENE_NORMAL][3]=0;
    g.bEbene=0;
    
    iFilterzeitMs=deltaT_ms(0);
    sollwertGesendetMs=deltaT_ms(0);
    b5msTic=0;
  }

  
  tastenPollen();
  
  //Ausgabewerte berechnen
  if (deltaT_ms(iFilterzeitMs) > 100){
    filterSchritt();
    iFilterzeitMs=deltaT_ms(0);
  }
  
  //Sollwerte ggf. senden
  b5msTic++;    
  if (40==b5msTic){ //200ms
    b5msTic=0;
    if ((deltaT_ms(sollwertGesendetMs) > 999) || (g.bSendeSollwerte>0)){
        sollwerteSenden();
        if  (g.bSendeSollwerte > 0) g.bSendeSollwerte--;
        sollwertGesendetMs=deltaT_ms(0);
    }
  }
  
//    LICHT = TASTE_2; 
//    SUMMER = DI_DCF77_BIT;
  
  // Uhren
  //--------------------------------------------------------------------------
  if (bMerkTic != g.bToggle1000ms){ //Eine Sekunde ist rum.
    bMerkTic = g.bToggle1000ms;
    
    
    g.bLichtCounter++;
    if (g.bLichtCounter>100) g.bLichtCounter=100;
    
    // Drehe die Uhr weiter.
    g.bSekunde++;
    if (g.bSekunde > 59){
      g.bSekunde = 0;
      g.bMinute++;
      if (g.bMinute > 59){
        g.bMinute = 0;
        g.bStunde++;
        if (g.bStunde > 23){
          g.bStunde = 0;
        }
      }
    }
    
    // Lass Timer einen Schritt laufen
    iTimerSecNext = 32400;   //9h
    for (i=0; i<4; i++){
      if (g.iCounter[EBENE_TIMER][i] > 0){
        g.iCounter[EBENE_TIMER][i]--;
        if (0==g.iCounter[EBENE_TIMER][i]){
          //Timer abgelaufen.
          //Platte aus
          g.iCounter[EBENE_NORMAL][i]=0;
          g.bRefreshDisplay=1;
          g.bBlinken=1;            
          //Summer ein
          if (g.bBeeperEnabled) beeperEin();
          g.bAbgelaufeneTimer|= (1<<i);

        }
        iTimerSec = g.iCounter[EBENE_TIMER][i];
        if (iTimerSecNext > iTimerSec){
          iTimerSecNext = iTimerSec;
        }
      }
    }
    g.iNaechsterTimer_s = iTimerSecNext;
    //Ofen Timer
    if (g.iCounter[EBENE_OFEN][3] > 0){
      g.iCounter[EBENE_OFEN][3]--;
      if (0==g.iCounter[EBENE_OFEN][3]){
        //Timer abgelaufen.
        //Ofen aus
        g.iCounter[EBENE_OFEN][1]=0;
        g.bRefreshDisplay=1;
        g.bBlinken=1;            
        //Summer ein
        if (g.bBeeperEnabled) beeperEin();
        g.bAbgelaufeneTimer|= 0xff; //Alle Bits ein

      }
      if (g.iCounter[EBENE_OFEN][3] < g.iNaechsterTimer_s){
        g.iNaechsterTimer_s = g.iCounter[EBENE_OFEN][3];
      }
    }

    // Schaltuhr
    //Schaltuhrwerte aus Drehgebern berechnen
    g.iSchaltuhrDauer=g.iCounter[EBENE_SCHALTUHR][0]*60 + g.iCounter[EBENE_SCHALTUHR][1];
    g.iSchaltuhrEnde=g.iCounter[EBENE_SCHALTUHR][2]*60 + g.iCounter[EBENE_SCHALTUHR][3];
    iSchaltuhrBeginn = g.iSchaltuhrEnde - g.iSchaltuhrDauer;
    if (iSchaltuhrBeginn < 0) iSchaltuhrBeginn +=1440; //ber Mitternacht
    if (g.iSchaltuhrDauer > 0){
      g.iCounter[EBENE_OFEN][3]=0;  //evtl. vorhandenen Timer lschen
      iJetzt = g.bStunde * 60 + g.bMinute;
      
      if (iSchaltuhrBeginn < g.iSchaltuhrEnde){ //Normalfall: ohne Tageswechsel
        if ((iJetzt >= iSchaltuhrBeginn) && (iJetzt < g.iSchaltuhrEnde)){
          g.bFreigabeOfen=1;
        }else{ 
          g.bFreigabeOfen=0;
        }
      }else{ //Sonderfall: Ofen ber Mitternacht
        if ((iJetzt >= iSchaltuhrBeginn) || (iJetzt < g.iSchaltuhrEnde)){
          g.bFreigabeOfen=1;
        }else{ 
          g.bFreigabeOfen=0;
        }
      }
      //Todo: DAUER nach Erreichen von Ende lschen. Dabei muss auch die Heizstufe gelscht werden (sonst geht der Ofen wieder an!).
    }else{
      g.bFreigabeOfen=1; //Ohne Schaltuhr: Ofen darf EIN
    }
        
    g.bRefreshDisplay=1; //sptestens nach 1 Sekunde
  }
  
  
  // Drehknoepfe
  //--------------------------------------------------------------------------
  switch(g.bEbene){
    case EBENE_NORMAL:     ebeneNormal();     break;
    case EBENE_TIMER:      ebeneTimer();      break;
    case EBENE_OFEN:       ebeneOfen();       break;
    case EBENE_SCHALTUHR:  ebeneSchaltuhr();  break;
    case EBENE_SERVICE:    ebeneService();    break;
  }
}
//-----------------------------------------------------------------------------
void presetGloblVar(void){
//-----------------------------------------------------------------------------
  U8 i;
  
  // Initialisierung der globalen Variablen
  //---------------------------------------------------------------------------
  g.bStunde=0;
  g.bMinute=0;
  g.bSekunde=0;

  g.bBeeperEnabled=1;
  
  g.bEbene=EBENE_NORMAL;
  for (i=0; i<4; i++){
    g.iMax      [EBENE_NORMAL][i]=14;
    g.iMin      [EBENE_NORMAL][i]=0;
    g.iModulo   [EBENE_NORMAL][i]=9999;
    g.iCounter  [EBENE_NORMAL][i]=0;
    g.iInc      [EBENE_NORMAL][i]=1;
    
    g.iMax      [EBENE_TIMER][i]=10800;  //3 Stunden
    g.iMin      [EBENE_TIMER][i]=0;
    g.iModulo   [EBENE_TIMER][i]=9999;
    g.iCounter  [EBENE_TIMER][i]=0;
    g.iInc      [EBENE_TIMER][i]=30;

  }
  //Kleinerer Wert fr berhhung berhht strker!
  g.iUeberhoehung[KNOPF_OL]=200;
  g.iUeberhoehung[KNOPF_OR]=100;  
  g.iUeberhoehung[KNOPF_UL]=200;
  g.iUeberhoehung[KNOPF_UR]=100;  
/*
       ====+====+====+====+
        180   Ofen  Ul+Oh 
        Kern 77    1:30:55
*/
  g.iMax    [EBENE_OFEN][0]=250;   //Solltemp
  g.iMin    [EBENE_OFEN][0]=50;
  g.iModulo [EBENE_OFEN][0]=9999;
  g.iCounter[EBENE_OFEN][0]=180;
  g.iInc    [EBENE_OFEN][0]=5;

  g.iMax    [EBENE_OFEN][1]=8;  //Heizart  
  g.iMin    [EBENE_OFEN][1]=0;
  g.iModulo [EBENE_OFEN][1]=9;
  g.iCounter[EBENE_OFEN][1]=0;
  g.iInc    [EBENE_OFEN][1]=1;

  g.iMax    [EBENE_OFEN][2]=95;   //Kerntemp
  g.iMin    [EBENE_OFEN][2]=30;
  g.iModulo [EBENE_OFEN][2]=9999;
  g.iCounter[EBENE_OFEN][2]=80;
  g.iInc    [EBENE_OFEN][2]=-1;

  g.iMax    [EBENE_OFEN][3]=10800;   //Ofen Timer (max. 3h)
  g.iMin    [EBENE_OFEN][3]=0;
  g.iModulo [EBENE_OFEN][3]=12999;
  g.iCounter[EBENE_OFEN][3]=0;
  g.iInc    [EBENE_OFEN][3]=60;



  g.iMax    [EBENE_SCHALTUHR][0]=4;   //Dauer (Stunden)
  g.iMin    [EBENE_SCHALTUHR][0]=0;
  g.iModulo [EBENE_SCHALTUHR][0]=9999;
  g.iCounter[EBENE_SCHALTUHR][0]=0;
  g.iInc    [EBENE_SCHALTUHR][0]=1;

  g.iMax    [EBENE_SCHALTUHR][1]=55;   //Dauer (Minuten)
  g.iMin    [EBENE_SCHALTUHR][1]=0;
  g.iModulo [EBENE_SCHALTUHR][1]=60;
  g.iCounter[EBENE_SCHALTUHR][1]=0;
  g.iInc    [EBENE_SCHALTUHR][1]=2;

  g.iMax    [EBENE_SCHALTUHR][2]=99;   //Endezeit (Stunden)
  g.iMin    [EBENE_SCHALTUHR][2]=0;
  g.iModulo [EBENE_SCHALTUHR][2]=24;
  g.iCounter[EBENE_SCHALTUHR][2]=8;
  g.iInc    [EBENE_SCHALTUHR][2]=1;

  g.iMax    [EBENE_SCHALTUHR][3]=55;   //Endezeit (Minuten)
  g.iMin    [EBENE_SCHALTUHR][3]=0;
  g.iModulo [EBENE_SCHALTUHR][3]=60;
  g.iCounter[EBENE_SCHALTUHR][3]=0;
  g.iInc    [EBENE_SCHALTUHR][3]=5;


  g.iMax    [EBENE_SERVICE][0]=4;   //Drehknpfe bisher ungenutzt
  g.iMin    [EBENE_SERVICE][0]=0;
  g.iModulo [EBENE_SERVICE][0]=9999;
  g.iCounter[EBENE_SERVICE][0]=0;
  g.iInc    [EBENE_SERVICE][0]=1;

  g.iMax    [EBENE_SERVICE][1]=4;   //
  g.iMin    [EBENE_SERVICE][1]=0;
  g.iModulo [EBENE_SERVICE][1]=9999;
  g.iCounter[EBENE_SERVICE][1]=0;
  g.iInc    [EBENE_SERVICE][1]=1;

  g.iMax    [EBENE_SERVICE][2]=4;   //
  g.iMin    [EBENE_SERVICE][2]=0;
  g.iModulo [EBENE_SERVICE][2]=9999;
  g.iCounter[EBENE_SERVICE][2]=0;
  g.iInc    [EBENE_SERVICE][2]=1;

  g.iMax    [EBENE_SERVICE][3]=4;   //
  g.iMin    [EBENE_SERVICE][3]=0;
  g.iModulo [EBENE_SERVICE][3]=9999;
  g.iCounter[EBENE_SERVICE][3]=0;
  g.iInc    [EBENE_SERVICE][3]=1;
  
};
//-----------------------------------------------------------------------------
void filterTemperaturen(void){  //alle 20ms
//-----------------------------------------------------------------------------
static I32 celsiusHerdFilter = 20*65536L;
static I32 celsiusKernFilter = 20*65536L;
  
  celsiusHerdFilter += ((I32)g.iOfentempNormiert * 65536L - celsiusHerdFilter) / 256;
  g.iOfentempCelsius = celsiusHerdFilter / 65536L;

  celsiusKernFilter += ((I32)g.iKerntempNormiert * 65536L - celsiusKernFilter) / 128;
  g.iKerntempCelsius = celsiusKernFilter / 65536L;

}
//-----------------------------------------------------------------------------
int main(void){
//-----------------------------------------------------------------------------
static I16 tic5ms=0;
static I16 ticDisplay=0;
static I16 ticBlinken=0;
static I16 ticBeepersweep=0;

  cli();
  initHW();
  USART_Init();

  // zur Initialisierung die Drehgeber aufrufen
  drehgeber(DREH1_A, DREH1_B, 0);
  drehgeber(DREH2_A, DREH2_B, 1);
  drehgeber(DREH3_A, DREH3_B, 2);
  drehgeber(DREH4_A, DREH4_B, 3);
  tastenPollen();

  presetGloblVar();
  _delay_ms(20);
  beeperAus();  
  ProcessLCDPanel();

  sei();
 
  while(1){

    //Alle 5ms den Ablauf aufrufen
    if (deltaT_ms(tic5ms) > 4){
      tic5ms = deltaT_ms(0);
      task_dcf77();  
      task_control();
    }

    //Alle 20ms das Display behandeln
    if (deltaT_ms(ticDisplay) > 19){
      ticDisplay = deltaT_ms(0);
      filterTemperaturen();
      ProcessLCDPanel();
      if (g.mindestPiepsTimer) g.mindestPiepsTimer--;

#if 0 //zum Debuggen
      g.bRefreshDisplay=1;
#endif
    }

    if (deltaT_ms(ticBeepersweep) > 999){ //Jede Sekunde den Piepston ndern
      ticBeepersweep = deltaT_ms(0);
      OCR2A--;
      if (OCR2A < 29){
        OCR2A = 89;
      }
    }

    //Empfangene Zeichen verarbeiten
    if (g.RxBufferWrite!=g.RxBufferRead){ //Zeichen vorhanden
      rxChar( g.RxBuffer[g.RxBufferRead++] ); 
      if (RX_BUFFERSIZE == g.RxBufferRead){
        g.RxBufferRead = 0;
      }
    }
      if (g.bBlinken){
        if (deltaT_ms(ticBlinken) > 100){
          ticBlinken = deltaT_ms(0);
          g.blinkphase = !g.blinkphase;
          g.bRefreshDisplay = 1;
        }
      }
  }//while(1)
}
//EOF
